import os
import sys
import datetime
import pytz
import logging
import random
import string

from kocher_tools.database import backupDatabase

class Backups (list):
	def __init__ (self, *arg, out_dir = None, limit = None, update_freq = None, **kw):
		super(Backups, self).__init__(*arg, **kw)
		self.out_dir = out_dir
		self.limit = limit
		self.update_freq = update_freq
		self.timezone = pytz.timezone('US/Eastern')
		self.str_format = '%Y-%m-%d'
		self.date = datetime.datetime.now(self.timezone)
		self.oldest_backup = None
		self.most_recent_backup = None

		# Check if data was not assigned
		if not self:

			# Assign the backups
			self.assignBackups()

	@property
	def date_str (self):

		# Return the date as a string of Year-Month-Day
		return self.date.strftime(self.str_format)

	def assignBackups (self):

		# Assign an empty object to store the oldest backup
		oldest_backup = None

		# Assign an empty object to store the newest backup
		most_recent_backup = None
        
		# Loop the files within the backup dir
		for backup_file in os.listdir(self.out_dir):

			# Assign the backup file path
			backup_file_path = os.path.join(self.out_dir, backup_file)

			# Assign the past backup date str
			past_backup_date_str = backup_file.rsplit('.', 2)[-2]

			# Assign the backup date
			past_backup_date = datetime.datetime.strptime(past_backup_date_str, self.str_format)

			# Update the past backup date with the timezone
			past_backup_date = past_backup_date.replace(tzinfo = self.timezone)

			# Assign a backup
			past_backup = Backup(backup_file_path, past_backup_date)

			# Check if the oldest backup has been defined
			if not oldest_backup:

				# Assign the current backup
				oldest_backup = past_backup

			else:

				# Check if the current backup is older than the stored oldest backup
				if past_backup < oldest_backup:

					# Assign the current backup
					oldest_backup = past_backup

			# Check if the newest backup has been defined
			if not most_recent_backup:

				# Assign the current backup
				most_recent_backup = past_backup

			else:

				# Check if the current backup is older than the stored oldest backup
				if past_backup > most_recent_backup:

					# Assign the current backup
					most_recent_backup = past_backup

			# Save the table
			self.append(past_backup)

		# Assing the oldest backup and the most recent date
		self.oldest_backup = oldest_backup
		self.most_recent_backup = most_recent_backup

	def updateBackups (self):

		# Check if the number of backups exceeds the backup limit
		if len(self) > self.limit:

			# Delete the oldest backup
			self.deleteBackup(self.oldest_backup)

			# Assign an empty object to store the oldest backup
			oldest_backup = None

			# Assign an empty object to store the newest backup
			most_recent_backup = None
	        
			# Loop the backups, backup by backup
			for current_backup in self:

				# Check if the oldest backup has been defined
				if not oldest_backup:

					# Assign the current backup
					oldest_backup = current_backup

				else:

					# Check if the current backup is older than the stored oldest backup
					if current_backup < oldest_backup:

						# Assign the current backup
						oldest_backup = current_backup

				# Check if the newest backup has been defined
				if not most_recent_backup:

					# Assign the current backup
					most_recent_backup = current_backup

				else:

					# Check if the current backup is older than the stored oldest backup
					if current_backup > most_recent_backup:

						# Assign the current backup
						most_recent_backup = current_backup

			# Assing the oldest backup and the most recent date
			self.oldest_backup = oldest_backup
			self.most_recent_backup = most_recent_backup

	def backupNeeded (self):

		# Check if there are no backups
		if not self.most_recent_backup:

			# Return True if their is no backup
			return True

		# Get the difference in time from the most recent backup
		date_diff = self.date - self.most_recent_backup.backup_date

		# Check if the difference in greater than the update frequency
		if self.update_freq >= date_diff.days:

			# Return False if the limit to larger than the number of days
			return False

		# Return True if the limit is smaller
		return True

	def newBackup (self, database):

		# Assign the database basename and file extension
		database_basename = os.path.basename(database)
			
		# Create random string for database filename
		random_str = ''.join(random.choice(string.ascii_uppercase + string.digits) for i in range(6))

		# Assign the backup filename
		backup_file = os.path.join(self.out_dir, '%s.%s.%s.backup' % (database_basename, random_str, self.date_str))

		# Create the database
		backupDatabase(database, backup_file)

		# Assign the new backup
		new_backup = Backup(backup_file, self.date)

		# Append the new backup
		self.append(new_backup)

		# Update the most recent backup
		self.most_recent_backup = new_backup

		# Update the backups
		self.updateBackups()

	def deleteBackup (self, backup_to_delete):

		# Assign the file path of the backup file
		backup_file_path = str(backup_to_delete)

		# Confirm the file exists, otherwise raise an error
		if not os.path.isfile(backup_file_path):
			raise Exception('Backup file (%s) not found' % backup_file_path)

		# Remove the file
		os.remove(backup_file_path)

		# Loop the backups by index
		for backup_pos in range(len(self)):

			# Check if the current backup is the one to remove
			if self[backup_pos] == backup_to_delete:

				# Remove the backup object
				del self[backup_pos]
				break

class Backup ():
	def __init__ (self, file_path, backup_date):

		# Confirm the file exists
		if not os.path.isfile(file_path):
			raise Exception('Backup file (%s) not found' % file_path)

		self.file_path = file_path
		self.backup_date = backup_date

	def __str__ (self):

		# str() returns the file path
		return self.file_path

	def __gt__ (self, other_backup):

		# Check if the stored date is more recent
		return self.backup_date > other_backup.backup_date

	def __lt__ (self, other_backup):

		# Check if the stored date is older
		return self.backup_date < other_backup.backup_date





import os
import yaml
import re

from collections import OrderedDict, defaultdict
from itertools import combinations

from sqlalchemy import Table, Column, ForeignKey, CheckConstraint, MetaData, Integer, Numeric, String, Text, LargeBinary, Date, DateTime, Boolean
from sqlalchemy.sql import func

class ConfigDB (list):
	def __init__ (self, yaml = ''):
		self.yaml = yaml
		self.yaml_dir = os.path.dirname(yaml)
		self.type = None
		self.filename = None
		self.host = None
		self.user = None
		self.passwd = None
		self.database = None
		self.schema = None
		self.meta = None
		self._sql_tables = {}
		self._table_label_to_col = defaultdict(lambda: defaultdict(str))
		self._table_col_to_label = defaultdict(lambda: defaultdict(str))

		# Assign using the Yaml file
		self.assignFromYaml()

	@property
	def tables (self):
		return list(self._sql_tables)

	@property
	def sql_tables (self):
		return list(self._sql_tables.values())

	@property
	def sql_address (self):
		if not self.user or not self.passwd or not self.host or not self.database:
			sql_address_err_list = []
			if not self.user: sql_address_err_list.append('SQL Username')
			if not self.passwd: sql_address_err_list.append('SQL Password')
			if not self.host: sql_address_err_list.append('SQL Host')
			if not self.database: sql_address_err_list.append('SQL Database')
			raise Exception('Unable to assign SQL Address. Missing: %s.' % ', '.join(sql_address_err_list))
		return '%s:%s@%s/%s' % (self.user, self.passwd, self.host, self.database)

	@property
	def passwd_required (self):
		if self.type == 'sqlite': return False
		else: return True

	def __contains__ (self, table_str):
		if table_str in self._sql_tables:
			return True
		else:
			return False

	def __getitem__(self, table_str):
		return self._sql_tables[table_str]
	
	@classmethod
	def readConfig (cls, yaml_filename):

		# Create a ConfigDB object
		return cls(yaml = yaml_filename)

	def getSQLTables (self, table_strs):

		# Confirm the tables exists, and if so, return the SQL tables
		tables_to_return = []
		for table_str in table_strs:
			if table_str not in self: raise Exception('Table (%s) not found' % table_str)
			tables_to_return.append(self[table_str])
		return tables_to_return

	@staticmethod
	def getSQLColumn (table, column_str):

		try: sql_column = getattr(table, 'unique_id')
		except: raise Exception('Unable to assign column: %s' % column_str)

		return sql_column

	'''
	def getSQLColumns (self, column_strs):

		# Confirm the columns exists, and if so, return the SQL columns
		columns_to_return = []
		for column_str in column_strs:
			column_arg = None
			column_assigned = False
			for sql_table in self._sql_tables.values():
				try:
					test_arg = getattr(sql_table.columns, column_str)
					if len(test_arg.foreign_keys) >= 1:
						continue
					column_assigned = True
					column_arg = test_arg
				except: pass
			if not column_assigned: raise Exception('Unable to assign column: %s' % column_str)
			columns_to_return.append(column_arg)
		return columns_to_return
	'''

	def assignFromYaml (self):

		# Read in the YAML config file
		with open(self.yaml, 'r') as yml_config_file:

			# Load the YAML config file in order
			config_yaml = yaml.load(yml_config_file, Loader=yaml.FullLoader)

			# Assign the database information
			self.type = os.path.join(config_yaml['sql']['type'])
			if self.type == 'sqlite':
				self.filename = os.path.join(config_yaml['sql']['filename'])
			else:
				self.host = os.path.join(config_yaml['sql']['host'])
				self.user = os.path.join(config_yaml['sql']['user'])
				if 'database' in config_yaml['sql']: self.database = os.path.join(config_yaml['sql']['database'])
				if 'schema' in config_yaml['sql']: self.schema = os.path.join(config_yaml['sql']['schema'])
				if config_yaml['sql']['passwd']: self.passwd = os.path.join(config_yaml['sql']['passwd'])

			# Create a metadata object
			if self.schema: meta = MetaData(schema = self.schema)
			else: meta = MetaData()

			# Loop the config yaml
			for table, column_yaml in config_yaml['database']['tables'].items():

				# Create the table attribute dict
				table_attr_list = []

				# Loop the config yaml
				for column, attribute_yaml in column_yaml.items():

					# Create a column arg list and kwarg dict
					col_arg_list = []
					col_kwarg_dict = {}

					# Assign a String-class arg, if found
					if 'String' in attribute_yaml['type']:
						if '(' in attribute_yaml['type'] and ')' in attribute_yaml['type']:
							type_value = int(attribute_yaml['type'].split('(')[1].split(')')[0])
							col_arg_list.append(String(type_value))
						else:
							col_arg_list.append(String)

					# Assign a Boolean-class arg, if found
					elif 'Boolean' in attribute_yaml['type']:
					
						col_arg_list.append(Boolean)
						
					# Assign a String-class arg, if found
					elif 'Text' in attribute_yaml['type']:
					
						col_arg_list.append(Text)

					# Assign a DateTime-class arg, if found
					elif 'DateTime' in attribute_yaml['type']:
						if 'default' in attribute_yaml or 'onupdate' in attribute_yaml: raise Exception('DateTime does not support default or onupdate options')
						if '(' in attribute_yaml['type'] and ')' in attribute_yaml['type']:
							if   '(record-created)': 
								col_arg_list.append(DateTime(timezone = True))
								col_kwarg_dict['default'] = func.now()
							elif '(record-updated)': 
								col_arg_list.append(DateTime(timezone = True))
								col_kwarg_dict['onupdate'] = func.now()
							elif    '(record-both)': 
								col_arg_list.append(DateTime(timezone = True))
								col_kwarg_dict['default'] = func.now()
								col_kwarg_dict['onupdate'] = func.now()
							else: raise Exception('Unable to assign DateTime subtype (%s)' % attribute_yaml['type'].split('(')[1].split(')')[0])
						else:
							col_arg_list.append(DateTime(timezone = True))

					# Assign a Date-class arg, if found
					elif 'Date' in attribute_yaml['type']:
						
						col_arg_list.append(Date)

					# Assign a Integer-class arg, if found
					elif 'Integer' in attribute_yaml['type']:
						
						col_arg_list.append(Integer)

					# Assign a Integer-class arg, if found
					elif 'Numeric' in attribute_yaml['type']:
						
						col_arg_list.append(Numeric)

					# Assign a Integer-class arg, if found
					elif 'Binary' in attribute_yaml['type']:
						
						col_arg_list.append(LargeBinary)

					else:
						if 'type' not in attribute_yaml: raise Exception('No type defined for column (%s)' % column)
						else: raise Exception('Unable to define type (%s) for column (%s)' % (attribute_yaml['type'], column))


					# Loop the optional column attributes
					for attribute_arg, attribute_value in attribute_yaml.items():

						# Assign as a primary_key, if found
						if attribute_arg == 'primary_key':
							col_kwarg_dict['primary_key'] = True

						# Assign as unique, if found
						if attribute_arg == 'unique':
							col_kwarg_dict['unique'] = True

						# Assign as not null, if found
						if attribute_arg == 'not_null':
							col_kwarg_dict['nullable'] = not attribute_value

						if attribute_arg == 'default':
							col_kwarg_dict['default'] = attribute_value

						if attribute_arg == 'onupdate':
							col_kwarg_dict['onupdate'] = attribute_value

						# Assign a ForeignKey arg, if found
						if attribute_arg == 'foreign_key':

							# Create foreign_key args
							fk_arg_list = []
							fk_kwarg_dict = {}

							for fk_arg, fk_value in attribute_value.items():
								if fk_arg == 'parent_key': fk_arg_list.append(fk_value)
								else: fk_kwarg_dict[fk_arg] = fk_value

							col_arg_list.append(ForeignKey(*fk_arg_list, **fk_kwarg_dict))

					# Add the column to the table attribute dict 
					sql_column = Column(column, *col_arg_list, **col_kwarg_dict)
					table_attr_list.append(sql_column)

					# Assign a CheckConstraint, if found, return error if constraint is unknown
					if 'constraint' in attribute_yaml:
						if attribute_yaml['constraint']['check'] == 'EQ':
							table_attr_list.append(CheckConstraint(sql_column == attribute_yaml['constraint']['value'], name = '%s_check' % column))
						elif attribute_yaml['constraint']['check'] == 'IN':
							table_attr_list.append(CheckConstraint(sql_column.in_(attribute_yaml['constraint']['value']), name = '%s_check' % column))
						elif attribute_yaml['constraint']['check'] == 'LT':
							table_attr_list.append(CheckConstraint(sql_column < attribute_yaml['constraint']['value'], name = '%s_check' % column))
						elif attribute_yaml['constraint']['check'] == 'GT':
							table_attr_list.append(CheckConstraint(sql_column > attribute_yaml['constraint']['value'], name = '%s_check' % column))
						elif attribute_yaml['constraint']['check'] == 'BOOL':
							table_attr_list.append(CheckConstraint(sql_column.in_([0, 1]), name = '%s_check' % column))
						elif attribute_yaml['constraint']['check'] == 'TF_1':
							table_attr_list.append(CheckConstraint(sql_column.in_(['T', 'F']), name = '%s_check' % column))
						elif attribute_yaml['constraint']['check'] == 'TF_Text':
							table_attr_list.append(CheckConstraint(sql_column.in_(['True', 'False']), name = '%s_check' % column))
						elif attribute_yaml['constraint']['check'] == 'YN':
							table_attr_list.append(CheckConstraint(sql_column.in_(['Y', 'N']), name = '%s_check' % column))
						elif attribute_yaml['constraint']['check'] == 'YN_NS':
							table_attr_list.append(CheckConstraint(sql_column.in_(['Y', 'N', 'NS']), name = '%s_check' % column))
						elif attribute_yaml['constraint']['check'] == 'YesNo':
							table_attr_list.append(CheckConstraint(sql_column.in_(['Yes', 'No']), name = '%s_check' % column))
						elif attribute_yaml['constraint']['check'] == 'YesNo_NS':
							table_attr_list.append(CheckConstraint(sql_column.in_(['Yes', 'No', 'NS']), name = '%s_check' % column))
						else: raise Exception('Unknown constraint (%s)' % attribute_yaml['constraint']['check'])

					# Assign a column label, if found
					if 'label' in attribute_yaml:
						self._table_label_to_col[table][attribute_yaml['label']] = column
						self._table_col_to_label[table][column] = attribute_yaml['label']

				# Assign the table class
				self._sql_tables[table] = Table(table, meta, *table_attr_list)

			self.meta = meta

	def webYaml (self):

		# Read in the YAML config file
		with open(self.yaml, 'r') as yml_config_file:

			# Load the YAML config file in order
			config_yaml = yaml.load(yml_config_file, Loader=yaml.FullLoader)

			# Create the yaml output dict and string
			yaml_output_dict = {}
			yaml_output_str = ''

			# Loop the config yaml
			for table, column_yaml in config_yaml['database']['tables'].items():

				# Populate the yaml output dict with the current table
				
				yaml_output_dict['form'] = f'{table}'
				yaml_output_dict['form_label'] = f'{table.capitalize()}'
				yaml_output_dict['db_schema'] = 'dash'
				yaml_output_dict['db_table'] = f'{table}'

				# Create the fields dict
				fields_dict = []

				# Loop the config yaml
				for column, attribute_yaml in column_yaml.items():

					# Create and populate the column field dict
					col_field_dict = {}

					col_field_dict['field'] = column
					col_field_dict['field_label'] =  attribute_yaml['label']

					# Assign a String-class arg, if found
					if 'String' in attribute_yaml['type']:

						col_field_dict['data_type'] = 'string'
						col_field_dict['data_class'] = 'text'
	
					# Assign a Boolean-class arg, if found
					elif 'Boolean' in attribute_yaml['type']:
						
						col_field_dict['data_type'] = 'boolean'
						col_field_dict['data_class'] = 'select'

					# Assign a String-class arg, if found
					elif 'Text' in attribute_yaml['type']:
					
						col_field_dict['data_type'] = 'string'
						col_field_dict['data_class'] = 'textarea'

					# Assign a Date-class arg or DateTime-class arg, if found
					elif 'DateTime' in attribute_yaml['type'] or 'Date' in attribute_yaml['type']:

						col_field_dict['data_type'] = 'string'
						col_field_dict['data_class'] = 'date'

					# Assign a Integer-class arg, if found
					elif 'Integer' in attribute_yaml['type']:
						
						col_field_dict['data_type'] = 'integer'
						col_field_dict['data_class'] = 'integer'

					# Assign a Integer-class arg, if found
					elif 'Numeric' in attribute_yaml['type']:
						
						col_field_dict['data_type'] = 'float'
						col_field_dict['data_class'] = 'real'

					else:
						if 'type' not in attribute_yaml: raise Exception('No type defined for column (%s)' % column)
						else: raise Exception('Unable to define type (%s) for column (%s)' % (attribute_yaml['type'], column))

					# Check if the column is the primary key
					if 'primary_key' in attribute_yaml:
						col_field_dict['primary_key'] = True

					# Add the column fields
					fields_dict.append(col_field_dict)

				# Add the fields to the yaml table dict
				yaml_output_dict['fields'] = fields_dict

				# Add to the yaml output string
				if yaml_output_str: yaml_output_str += '\n'
				yaml_output_str += yaml.dump(yaml_output_dict, explicit_start = True, sort_keys = False, indent = 4)

			# Modify the yaml output string
			yaml_output_str = re.sub(r'([a-z])\n\-  ', r'\1\n\n-  ', yaml_output_str)
			yaml_output_str = yaml_output_str.replace('-   ', '  - ')
			return yaml_output_str

	def webYamlFile (self, yaml_filename):
		if not yaml_filename: raise Exception('No filename specified')

		web_yaml_file = open(yaml_filename, 'w')
		web_yaml_file.write(self.webYaml())
		web_yaml_file.close()

	def webValueYaml (self):

		# Read in the YAML config file
		with open(self.yaml, 'r') as yml_config_file:

			# Load the YAML config file in order
			config_yaml = yaml.load(yml_config_file, Loader=yaml.FullLoader)

			# Create list to store previous value lists, to avoid repeats
			valuelist_list = []

			# Create a valuelist counter
			valuelist_counter = 1

			# Create the yaml out string
			yaml_output_str = ''

			# Loop the config yaml
			for table, column_yaml in config_yaml['database']['tables'].items():
				
				# Loop the config yaml
				for column, attribute_yaml in column_yaml.items():

					# Create the yaml output dict
					yaml_output_dict = {}
					
					# Check if the column has a constraint
					if 'constraint' in attribute_yaml:
						if 'value' in attribute_yaml['constraint']:

							constraint_type = attribute_yaml['constraint']['check']
							value_list = attribute_yaml['constraint']['value']

							# Create and populate the constraint options dict
							constraint_options_dict = {}
							constraint_options_dict['opt_type'] = 'array'
							if constraint_type == 'IN': pass
							else: raise Exception(attribute_yaml['constraint'])

							constraint_options_dict['opt_values'] = value_list
							yaml_output_dict['valuelist'] = f'{constraint_type}_{valuelist_counter}'
							yaml_output_dict['options'] = constraint_options_dict

							yaml_output_str +=  yaml.dump(yaml_output_dict, explicit_start = True, sort_keys = False, indent = 4, )
							yaml_output_str = yaml_output_str.replace("'", '')
							valuelist_list.append(value_list)
							valuelist_counter += 1

						else:
							constraint_type = attribute_yaml['constraint']['check']
							
							# Create and populate the constraint options dict
							constraint_options_dict = {}
							constraint_options_dict['opt_type'] = 'array'
							if constraint_type == 'YesNo_NS': value_list = ['Yes', 'No', 'NS']
							elif constraint_type == 'YesNo': value_list = ['Yes', 'No']
							else: raise Exception(constraint_type)
							constraint_options_dict['opt_values'] = value_list
							yaml_output_dict['valuelist'] = f'{constraint_type}'
							yaml_output_dict['options'] = constraint_options_dict

							yaml_output_str +=  yaml.dump(yaml_output_dict, explicit_start = True, sort_keys = False, indent = 4, )
							yaml_output_str = yaml_output_str.replace("'", '')
							valuelist_list.append(constraint_type)

			return yaml_output_str



import os
import sys
import csv
import logging

from collections import OrderedDict 

def unquoteFields (db_entries):

	# Create list to hold unquoted data
	unquoted_db_entries = []

	# Loop the entries
	for db_entry in db_entries:

		# Create an empty dict
		db_entry_dict = OrderedDict()

		# Loop the headers and the values
		for column, value in zip(db_entry.keys(), db_entry):

			# Convert the data to string and unquote
			db_entry_dict[str(column).replace('"','')] = str(value).replace('"','')

		# Append the dict to the list
		unquoted_db_entries.append(db_entry_dict)

	return unquoted_db_entries

def entriesToScreen (db_entries, sep):

	# Unquote the DB entries
	unquoted_db_entries = unquoteFields(db_entries)

	# Get the header from the first entry
	header = unquoted_db_entries[0].keys()

	# Print the headers to stdout, seperated by sep
	print(sep.join(header))

	# Loop each database entry
	for db_entry_dict in unquoted_db_entries:

		# Print the values to stdout, seperated by sep
		print(sep.join(db_entry_dict.values()))

	# Update log
	logging.info('Retrieved results sent to stdout')

def entriesToFile (db_entries, out_filename, sep):

	# Unquote the DB entries
	unquoted_db_entries = unquoteFields(db_entries)

	# Get the header from the first entry
	header = unquoted_db_entries[0].keys()

	# Open the file
	with open(out_filename, 'w') as entries_file:

		# Setup the csv writer
		entries_writer = csv.DictWriter(entries_file, fieldnames = header, delimiter = sep)

		# Write the header
		entries_writer.writeheader()

		# Loop each database entry
		for db_entry_dict in unquoted_db_entries:

			# Write the database entry to the file
			entries_writer.writerow(db_entry_dict)

	# Update log
	logging.info('Retrieved results written to file (%s)' % out_filename)



import os
import sys
import argparse
import logging
import sqlite3

from collections import defaultdict

from kocher_tools.output import entriesToScreen, entriesToFile
from kocher_tools.assignment import assignSelectionDict, assignTables
from kocher_tools.database import retrieveValues
from kocher_tools.config_file import readConfig
from kocher_tools.logger import startLogger, logArgs

def retrieveSampleParser ():
	'''
	Database Argument Parser 

	Raises
	------
	IOError
		If the input, or other specified files do not exist
	'''

	def confirmFile ():
		'''Custom action to confirm file exists'''
		class customAction(argparse.Action):
			def __call__(self, parser, args, value, option_string=None):
				if not os.path.isfile(value):
					raise IOError('%s not found' % value)
				setattr(args, self.dest, value)
		return customAction

	def selectionList ():
		'''Custom action to add items to a list'''
		class customAction(argparse.Action):
			def __call__(self, parser, args, value_list, option_string=None):

				# Clean up any commas
				value_list = [item.strip(',') for item in value_list]

				if not getattr(args, self.dest):
					setattr(args, self.dest, value_list)
				else:
					getattr(args, self.dest).extend(value_list)
		return customAction

	def selectionDict ():
		'''Custom action to add items to a list'''
		class customAction(argparse.Action):
			def __call__(self, parser, args, value_list, option_string=None):

				# Clean up any commas
				value_list = [item.strip(',') for item in value_list]

				if not getattr(args, self.dest):
					# Create the dict
					value_dict = defaultdict(list)
					# Populate the dict
					value_dict[value_list[0]].append(value_list[1])
					setattr(args, self.dest, value_dict)
				else:
					getattr(args, self.dest)[value_list[0]].append(value_list[1])
		return customAction

	def metavarList (var_list):
		'''Create a formmated metavar list for the help output'''
		return '{' + ', '.join(var_list) + '}'

	retrieve_parser = argparse.ArgumentParser(formatter_class = argparse.ArgumentDefaultsHelpFormatter)

	# Selection arguments
	retrieve_select = retrieve_parser.add_mutually_exclusive_group(required = True)
	retrieve_select.add_argument('--table', help = 'Table(s) to return', type = str, nargs = '+', action = selectionList())
	retrieve_select.add_argument('--column', help = 'Column(s) to return', type = str, nargs = '+', action = selectionList())
	retrieve_parser.add_argument('--include-ID', help = 'ID to include in database retrievals', type = str, nargs = '+', action = selectionList())
	retrieve_parser.add_argument('--exclude-ID', help = 'ID to exclude in database retrievals', type = str, nargs = '+', action = selectionList())
	retrieve_parser.add_argument('--include-species', help = 'Species to include in database retrievals', type = str, nargs = '+', action = selectionList())
	retrieve_parser.add_argument('--exclude-species', help = 'Species to exclude in database retrievals', type = str, nargs = '+', action = selectionList())
	retrieve_parser.add_argument('--include-genus', help = 'Genus to include in database retrievals', type = str, nargs = '+', action = selectionList())
	retrieve_parser.add_argument('--exclude-genus', help = 'Genus to exclude in database retrievals', type = str, nargs = '+', action = selectionList())
	retrieve_parser.add_argument('--include-nests', help = 'Include samples from nests in database retrievals', action = 'store_true')
	retrieve_parser.add_argument('--exclude-nests', help = 'Exclude samples from nests in database retrievals', action = 'store_true')
	retrieve_parser.add_argument('--include', metavar = ('column', 'value'), help = 'Column/value pair to include in database retrievals', type = str, nargs = 2, action = selectionDict())
	retrieve_parser.add_argument('--exclude', metavar = ('column', 'value'), help = 'Column/value pair to exclude in database retrievals', type = str, nargs = 2, action = selectionDict())

	# Output arguments
	out_formats = ['tsv', 'csv']
	out_default = 'tsv'
	retrieve_parser.add_argument('--out-format', metavar = metavarList(out_formats), help = 'Desired output format', type = str, choices = out_formats, default = out_default)
	retrieve_parser.add_argument('--out-prefix', help = 'The output prefix (i.e. filename without file extension)', type = str,  default = 'out')
	retrieve_parser.add_argument('--out', help = 'Desired output name. Overrides --out-prefix', type = str)
	retrieve_parser.add_argument('--out-columns', help = 'Columns to return in the output', type = str, nargs = '+', action = selectionList())
	retrieve_parser.add_argument('--out-log', help = 'Filename of the log file', type = str, default = 'retrieve_samples.log')
	retrieve_stdout = retrieve_parser.add_mutually_exclusive_group()
	retrieve_stdout.add_argument('--stdout', help = 'Direct output to stdout', action = 'store_true')
	retrieve_stdout.add_argument('--log-stdout', help = 'Direct logging to stdout', action = 'store_true')
	retrieve_parser.add_argument('--overwrite', help = 'Overwrite previous output', action = 'store_true')

	# Database arguments
	retrieve_parser.add_argument('--sqlite-db', help = 'Defines the sqlite database filename', type = str, required = True, action = confirmFile())
	retrieve_parser.add_argument('--yaml', help = 'Database YAML config file', type = str, required = True, action = confirmFile())

	return retrieve_parser.parse_args()

def main():
	
	# Assign arguments
	retrieve_args = retrieveSampleParser()

	# Create string to hold the output filename
	out_filename = ''

	# Check if an output filename was specified
	if retrieve_args.out:

		# Assign the out filename
		out_filename = retrieve_args.out

	# Assign the filename with --out-prefix otherwise
	else:

		# Assign the out filename
		out_filename = retrieve_args.out_prefix  + '.' + retrieve_args.out_format

	# Check that the output mode isn't stdout
	if not retrieve_args.stdout:

		# Check if previous output should be overwritten
		if retrieve_args.overwrite:

			# Remove the previous output, if it exists
			if os.path.exists(out_filename):
				os.remove(out_filename)

		# Check if previous output shouldn't be overwritten
		else:

			# Check if previous output exists
			if os.path.exists(out_filename):

				# Raise an exception
				raise Exception('Output already exists. Please alter the output arguments or use --overwrite')

	# Check if log should be sent to stdout
	if retrieve_args.log_stdout:

		# Start logging to stdout for this run
		startLogger()

	else:

		# Start a log file for this run
		startLogger(log_filename = retrieve_args.out_log)

	# Read in the database config file
	db_config_data = readConfig(retrieve_args.yaml)

	# Log the arguments used
	logArgs(retrieve_args)

	# List to hold columns to be printed
	columns_assignment_list = []

	# List to hold table to be printed, required for printing single columns
	table_assignment_list = []

	# Check if the user specified a table
	if retrieve_args.table:

		# Assign the tables
		table_assignment_list.extend(retrieve_args.table)

		# Loop the list of specified tables
		for table_str in table_assignment_list:

			# Check if the database has the specified table
			if table_str not in db_config_data:

				# Print an error message
				raise Exception('Table (%s) not found. Please select from: %s' % (table_str, ', '.join(db_config_data.tables)))

			# Get the columns from the table and add to the column assignment list
			columns_assignment_list.extend(db_config_data[table_str].retieveColumnPaths(keep_db_specific_columns = True))

			# Update log
			logging.info('Successfully assigned table(s) from command-line')

	# Check if the user specified a column
	elif retrieve_args.column:
		
		# Assign the tables
		table_assignment_list.extend(db_config_data.returnTables(retrieve_args.column))

		# Loop the list of specified column
		for column_str in retrieve_args.column:

			# Assign the column path
			columns_assignment_list.append(db_config_data.returnColumnPath(column_str))

		# Update log
		logging.info('Successfully assigned column(s) from command-line')

	# Assign a defaultdict with all the selection information
	selection_dict = assignSelectionDict(db_config_data, **vars(retrieve_args))

	# Assign the tables that need to be updated
	selection_tables = assignTables(db_config_data, **vars(retrieve_args))

	# Assign the tables for retrieval
	tables = table_assignment_list + selection_tables

	# Remove potential duplicates
	tables = list(set(tables))

	# Connect to the sqlite database
	sqlite_connection = sqlite3.connect(retrieve_args.sqlite_db)

	# Setup SQLite to reture the rows as dict with columns
	sqlite_connection.row_factory = sqlite3.Row

	# Create the cursor
	cursor = sqlite_connection.cursor()

	# Check if only a single table is required
	if len(tables) == 1:

		# Retrieve the selected entries from the database 
		retrieved_entries = retrieveValues(cursor, tables, selection_dict, columns_assignment_list)

	# Otherwise, run the process for multiple tables
	else:

		# Assign the tables and keys that need to be joined
		tables, join_by_columns = db_config_data.returnJoinLists(tables)

		# Retrieve the selected entries from the database 
		retrieved_entries = retrieveValues(cursor, tables, selection_dict, columns_assignment_list, join_table_columns = join_by_columns)

	# Commit any changes
	sqlite_connection.commit()

	# Close the connection
	cursor.close()

	# Set the deafult delimiter
	delimiter = '\t'

	# Check if the csv format was requested
	if retrieve_args.out_format == 'csv':

		# Update the delimiter
		delimiter = ','

	# Check if stdout was requested
	if retrieve_args.stdout:

		# Print the entries to stdout
		entriesToScreen(retrieved_entries, delimiter)

	else:

		# Write retrieved entries to a file
		entriesToFile(retrieved_entries, out_filename, delimiter)

if __name__ == "__main__":
	main()


import os
import sys
import gzip
import argparse

from Bio import SeqIO
from collections import defaultdict

def nStatArgs ():

	'''
	Argument parser

	Raises
	------
	IOError
		If the input, or other specified files do not exist
	'''

	def confirmFile ():
		'''Custom action to confirm file exists'''
		class customAction(argparse.Action):
			def __call__(self, parser, args, value, option_string=None):
				if not os.path.isfile(value):
					raise IOError('%s not found' % value)
				setattr(args, self.dest, value)
		return customAction

	nstat_parser = argparse.ArgumentParser(formatter_class = argparse.ArgumentDefaultsHelpFormatter)

	# Assign I/O arguments
	nstat_parser.add_argument('--genome-fasta', help = "Defines the genome filename (in fasta format)", type = str, action = confirmFile(), required = True)
	nstat_parser.add_argument('--output', help = "Defines the output n-statistics filename", type = str, required = True)

	# Return arguments
	return nstat_parser.parse_args()

def fastaStats (fasta_filename):

	# Read in the fasta file
	if fasta_filename.endswith('.gz'): fasta_file = gzip.open(fasta_filename, 'rt')
	else: fasta_file = open(fasta_filename, 'r')
	
	# Record the length for each sequence - i.e. chromosome, scaffold
	seq_lens = []
	n_count = 0
	for seq_record in SeqIO.parse(fasta_file, 'fasta'):
		seq_lens.append(len(seq_record))
		n_count += seq_record.seq.count('N')

	# Sort and return
	seq_lens = sorted(seq_lens, reverse = True)
	return seq_lens, n_count

def nStats (len_list, n_floats = [0.1, 0.5, 0.9], n_float = None):

	# Check if a  n_float was given
	if n_float: n_floats = [n_float]

	# Check the floats
	for _nf in n_floats: 
		if not isinstance(_nf, float): raise Exception(f'N calc must be float: {_nf}')

	# Obtain the total genome length and %n length
	genome_length = sum(len_list)
	n_cutoff_dict = {(genome_length * _nf):f'N{int((_nf) * 100)}'for _nf in n_floats}
	n_min_cutoff = min(list(n_cutoff_dict))
	
	# Calc the n stat
	n_stats = ''
	n_sum = 0
	n_count = 0
	for len_int in len_list:
		n_sum += len_int
		n_count += 1
		if n_sum >= n_min_cutoff: 
			n_stats += f'{n_cutoff_dict[n_min_cutoff]} = {len_int}, n = {n_count}\n'
			del n_cutoff_dict[n_min_cutoff]
			if not n_cutoff_dict: break
			else: n_min_cutoff = min(list(n_cutoff_dict))
	
	# Return the results
	return n_stats

# Assign the arguments using argparse
nStat_args = nStatArgs()

# Calc the sequence stats
scaffold_lens, n_count = fastaStats(nStat_args.genome_fasta)

# Create the output stat file
n_stat_output = open(nStat_args.output, 'w')
n_stat_output.write(f'stats for {nStat_args.genome_fasta}\n')
n_stat_output.write(f'sum = {sum(scaffold_lens)}, n = {len(scaffold_lens)}, ave = {sum(scaffold_lens)/len(scaffold_lens)}, largest = {max(scaffold_lens)}\n')
n_stat_output.write(nStats(scaffold_lens))
n_stat_output.write(f'N_count = {n_count}\n')
n_stat_output.close()


import os
import sys
import logging
import subprocess

from kocher_tools.misc import confirmExecutable

def checkVsearchForErrors (vsearch_stderr):

	# Check if an error was reported in the stderr
	if 'Fatal error' in vsearch_stderr:

		# Report the error
		raise Exception(vsearch_stderr)

def callVsearch (vsearch_call_args):

	# Find the vsearch executable
	vsearch_executable = confirmExecutable('vsearch')

	# Check if executable is installed
	if not vsearch_executable:
		raise IOError('vsearch not found. Please confirm the executable is installed')

	# vsearch subprocess call
	vsearch_call = subprocess.Popen([vsearch_executable] + vsearch_call_args, stderr = subprocess.PIPE, stdout = subprocess.PIPE)

	# Get stdout and stderr from subprocess
	vsearch_stdout, vsearch_stderr = vsearch_call.communicate()

	# Check if code is running in python 3
	if sys.version_info[0] == 3:
		
		# Convert bytes to string
		vsearch_stdout = vsearch_stdout.decode()
		vsearch_stderr = vsearch_stderr.decode()

	# Check for errors
	checkVsearchForErrors(vsearch_stderr)

def mergePairs (sample, r1_file, r2_file, merged_file, unmerged_R1_file, unmerged_R2_file, min_merge_len = 360, max_diff = 20):

	# Assign the R1 and R2 files
	merge_pairs_args = ['--fastq_mergepairs', r1_file, '--reverse', r2_file]

	# Assign the merged output file
	merge_pairs_args.extend(['--fastqout', merged_file])

	# Relabel with the sample name
	merge_pairs_args.extend(['--relabel', '%s_' % sample])

	# Assign the unmerged files
	merge_pairs_args.extend(['--fastqout_notmerged_fwd', unmerged_R1_file, '--fastqout_notmerged_rev', unmerged_R2_file])

	# Assign the quality control args
	merge_pairs_args.extend(['--fastq_minmergelen', str(min_merge_len), '--fastq_maxdiffs', str(max_diff)])

	# Call VSEARCH
	callVsearch(merge_pairs_args)

def truncateFastq (input_file, truncated_file, strip_left_n = 26, strip_right_n = 26):

	# Assign the input file
	truncate_args = ['--fastq_filter', input_file]

	# Assign the truncated output file
	truncate_args.extend(['--fastqout', truncated_file])

	# Assign the truncate args
	truncate_args.extend(['--fastq_stripleft',  str(strip_left_n), '--fastq_stripright', str(strip_right_n)])

	# Call VSEARCH
	callVsearch(truncate_args)

def filterFastq (input_file, filtered_file, maxee_float = 0.5):

	# Assign the input file
	filter_args = ['--fastq_filter', input_file]

	# Assign the filtered output file
	filter_args.extend(['--fastaout', filtered_file])

	# Assign the filter args
	filter_args.extend(['--fastq_maxee', str(maxee_float)])

	# Call VSEARCH
	callVsearch(filter_args)

def dereplicateFasta (plate, sample, input_file, dereplicated_file, min_unique_int = 2):

	# Assign the input file
	dereplicate_args = ['--derep_fulllength', input_file]

	# Assign the dereplicated output file
	dereplicate_args.extend(['--output', dereplicated_file])

	# Relabel with the sample name
	dereplicate_args.extend(['--relabel', '%s-%s_' % (plate, sample)])

	# Assign the dereplicate args
	dereplicate_args.extend(['--minuniquesize', str(min_unique_int), '--sizeout'])

	# Call VSEARCH
	callVsearch(dereplicate_args)

def clusterFasta (plate, sample, input_file, clustered_file, id_float = 1.0):

    # Assign the input file
	cluster_args = ['--cluster_smallmem', input_file]

	# Assign the clustered output file
	cluster_args.extend(['--centroids', clustered_file])

	# Relabel with the sample name
	cluster_args.extend(['--relabel', '%s-%s_' % (plate, sample)])

	# Assign the clustering args
	cluster_args.extend(['--id', str(id_float), '--sizein', '--sizeout', '--usersort'])

	# Call VSEARCH
	callVsearch(cluster_args)

def sortFasta (plate, sample, input_file, sorted_file):

    # Assign the input file
	sort_args = ['--sortbysize', input_file]

	# Assign the sorted output file
	sort_args.extend(['--output', sorted_file])

	# Relabel with the sample name
	sort_args.extend(['--relabel', '%s-%s_' % (plate, sample)])

	# Assign the sort args
	sort_args.extend(['--sizeout'])

	# Call VSEARCH
	callVsearch(sort_args)




#!/usr/bin/env python
import os
import sys
import argparse
import shutil
import logging
import multiprocessing
import pkg_resources

from kocher_tools.multiplex import Multiplex
from kocher_tools.blast import blastTopHits
from kocher_tools.logger import startLogger, logArgs

def barcodePipelineParser ():
	'''
	Barcode Pipeline Parser

	Assign the parameters for the barcode pipeline

	Parameters
	----------
	sys.argv : list
		Parameters from command lind

	Raises
	------
	IOError
		If the specified files do not exist
	'''

	def parser_confirm_file ():
		'''Custom action to confirm file exists'''
		class customAction(argparse.Action):
			def __call__(self, parser, args, value, option_string=None):
				if not os.path.isfile(value):
					raise IOError('%s not found' % value)
				setattr(args, self.dest, value)
		return customAction

	pipeline_parser = argparse.ArgumentParser(formatter_class = argparse.ArgumentDefaultsHelpFormatter)

	# Map files
	pipeline_parser.add_argument('--i5-map', help = 'Defines the filename of the i5 map', type = str, action = parser_confirm_file(), required = True)
	pipeline_parser.add_argument('--i7-map', help = 'Defines the filename of the i7 map (if not the default map)', type = str, action = parser_confirm_file())

	# Read Files
	pipeline_parser.add_argument('--i5-read-file', help = 'Defines the filename of the i5 reads (i.e. Read 3 Index)', type = str, action = parser_confirm_file(), required = True)
	pipeline_parser.add_argument('--i7-read-file', help = 'Defines the filename of the i7 reads (i.e. Read 2 Index)', type = str, action = parser_confirm_file(), required = True)
	pipeline_parser.add_argument('--R1-read-file', help = 'Defines the filename of the R1 reads (i.e. Read 1)', type = str, action = parser_confirm_file(), required = True)
	pipeline_parser.add_argument('--R2-read-file', help = 'Defines the filename of the R2 reads (i.e. Read 4)', type = str, action = parser_confirm_file(), required = True)

	# Output arguments
	pipeline_parser.add_argument('--out-dir', help = 'Defines the output directory', type = str, default = 'Pipeline_Output')
	pipeline_parser.add_argument('--out-compiled', help = 'Defines the filename of the compiled reads (i.e. most abundant)', type = str, default = 'Common.fasta')
	pipeline_parser.add_argument('--out-blast', help = 'Defines the filename of the BLAST output', type = str, default = 'BLAST.out')
	pipeline_parser.add_argument('--out-log', help = 'Defines the filename of the log file', type = str, default = 'barcode_pipeline.log')
	pipeline_parser.add_argument('--overwrite', help = 'Defines if previous output should be overwritten', action = 'store_true')

	# Optional arguments
	pipeline_parser.add_argument('--threads', help = 'Defines the number of threads. Default is all available threads', type = int, default = multiprocessing.cpu_count())
	pipeline_parser.add_argument('--blast-database', help = 'Defines the blast database. Default is the COI database', type = str, default = 'BLAST_DBs/Filtered_BOLD.fasta')

	# Return the arguments
	return pipeline_parser.parse_args()

def main():

	# Assign the barcode args
	barcode_args = barcodePipelineParser()
	
	# Check if no i7 map was assigned
	if not barcode_args.i7_map:
		
		# Assign the i7 map path from the package
		i7_map_path = pkg_resources.resource_filename('kocher_tools', 'data/i7_map.txt')
		
		# Check if i7 map does not exists
		if not os.path.exists(i7_map_path):
			
			# Return an error
			raise IOError('Cannot assign i7 map from package')
		
		# Assign the i7 map
		barcode_args.i7_map = i7_map_path

	# Create the log file
	startLogger(barcode_args.out_log)
	
	# Log the arguments used
	logArgs(barcode_args)

	# Check for previous output
	if os.path.exists(barcode_args.out_dir):

		# Check if previous output should be overwritten
		if barcode_args.overwrite:

			# Remove the previous output
			shutil.rmtree(barcode_args.out_dir) 

		# Check if previous output shouldn't be overwritten
		else: 

			# Raise an exception
			raise Exception('%s already exists. Please alter --out-dir or use --overwrite' % barcode_args.out_dir)

	# Create the multiplex job
	demultiplex_job = Multiplex()

	# Assign the output path for all multiplex files
	demultiplex_job.assignOutputPath(barcode_args.out_dir)

	# Assign the read file for the multiplex job
	demultiplex_job.assignFiles(barcode_args.i5_read_file, barcode_args.i7_read_file, barcode_args.R1_read_file, barcode_args.R2_read_file)

	# Assign the plate using the i5 map
	demultiplex_job.assignPlates(barcode_args.i5_map)

	logging.info('Starting i5 deMultiplex')

	# Run the i5 barcode job using the i5 map
	demultiplex_job.deMultiplex(barcode_args.i5_map)

	logging.info('Finished i5 deMultiplex')

	# Move the plates into directories of their plate and locus names 
	demultiplex_job.movePlates()

	# Remove unmatched files (this should be an option in beta)
	demultiplex_job.removeUnmatched()

	# Loop the plates in the multiplex job
	for plate in demultiplex_job:

		# Assign the well of the current plate
		plate.assignWells()

		logging.info('Starting %s i7 deMultiplex' % plate.name)

		# Run the i7 barcode job using the i7 map
		plate.deMultiplexPlate(barcode_args.i7_map)

		logging.info('Finished %s i7 deMultiplex' % plate.name)

		# Move the wells into the Wells directory. Should this be user specified?
		plate.moveWells()

		# Remove any unmatched files for the current plate
		plate.removeUnmatchedPlate()

		logging.info('Starting %s abundant reads identification' % plate.name)

		# Loop each well
		for well in plate:

			# Merge the R1/R2 files for the current well
			well.mergeWell()

			# Truncate the merged file for the current well
			well.truncateWell()

			# Filter the truncated file for the current well
			well.filterWell()

			# Dereplicate the filtered file for the current well
			well.dereplicateWell()

			# Cluster the dereplicated file for the current well
			well.clusterWell()

			# Identify the most abundant reads
			well.mostAbundantWell()

		logging.info('Finished %s abundant reads identification' % plate.name)

	# Assign the compiled file path
	compiled_file_path = os.path.join(barcode_args.out_dir, barcode_args.out_compiled)

	# Assign the blast output file path
	blast_file_path = os.path.join(barcode_args.out_dir, barcode_args.out_blast)

	# Compile the most abundant reads into a single file
	demultiplex_job.compileMostAbundant(compiled_file_path)

	logging.info('Starting BLAST')

	# Get the top BLAST hits for each sequence in the compiled file
	blastTopHits(compiled_file_path, blast_file_path, barcode_args.blast_database, barcode_args.threads)

	logging.info('Finished BLAST')

if __name__== "__main__":
	main()


import logging

import pandas as pd
#from collections import defaultdict

def readInputFile (input_file):

	# Create an empty dataframe
	input_datafreame = pd.DataFrame()

	# Try to read an excel input file
	try: input_datafreame = pd.read_excel(input_file, dtype = str)
	except:

		# Try to read a tsv input file
		try: input_datafreame = pd.read_csv(input_file, dtype = str, sep = '\t')
		except:

			# Try to read a csv input file
			try: input_datafreame = pd.read_csv(input_file, dtype = str)
			except: pass

	# Return an exception if the input is empty
	if input_datafreame.empty: raise Exception(f'Unable to parse input file: {input_file}')

	# Return the dataframe if not empty
	logging.info(f'Successfully assigned input file: {input_file}')
	return input_datafreame


'''
def assignTables (config_data, include = None, exclude = None, include_ID = None, exclude_ID = None, include_species = None, exclude_species = None, include_genus = None, exclude_genus = None, include_nests = None, exclude_nests = None, **kwargs):

	# Create a list to hold the columns
	columns = []
	
	# Check if basic include is defined
	if include:

		# Update the column list
		columns.extend(list(include.keys()))

	# Check if basic exclude is defined
	if exclude:

		# Update the column list
		columns.extend(list(exclude.keys()))

	# Check if ID include is defined
	if include_ID:

		# Update the column list
		columns.append('Unique ID')

	# Check if ID exclude is defined
	if exclude_ID:

		# Update the column list
		columns.append('Unique ID')

	# Check if species include is defined
	if include_species:

		# Update the column list
		columns.append('Species')

	# Check if species exclude is defined
	if exclude_species:

		# Update the column list
		columns.append('Species')

	# Check if genus include is defined
	if include_genus:

		# Update the column list
		columns.append('Species')

	# Check if genus exclude is defined
	if exclude_genus:

		# Update the column list
		columns.append('Species')

	# Check if nests include is defined
	if include_nests:

		# Update the column list
		columns.append('From Nest?')

	# Check if genus exclude is defined
	if exclude_nests:

		# Update the column list
		columns.append('From Nest?')


	# Assign the tables
	assigned_tables = config_data.returnTables(columns)

	# Update log
	logging.info('Successfully assigned selection statement table(s)')

	# Return the tables
	return assigned_tables
	
def assignSelectionDict (config_data, include = None, exclude = None, include_ID = None, exclude_ID = None, include_species = None, exclude_species = None, include_genus = None, exclude_genus = None,include_nests = None, exclude_nests = None,  **kwargs):

	# Create a defaultdict to hold all the selection information
	selection_dict = defaultdict(lambda: defaultdict(list))

	# Check if basic include is defined
	if include:

		# Update the selection dict
		selection_dict['IN'] = config_data.returnColumnPathDict(include)

	# Check if basic exclude is defined
	if exclude:

		# Update the selection dict
		selection_dict['NOT IN'] = config_data.returnColumnPathDict(exclude)

	# Check if ID include is defined
	if include_ID:

		# Update the selection dict
		selection_dict['IN'][config_data.returnColumnPath('Unique ID')] = include_ID

	# Check if ID exclude is defined
	if exclude_ID:

		# Update the selection dict
		selection_dict['NOT IN'][config_data.returnColumnPath('Unique ID')] = exclude_ID

	# Check if species include is defined
	if include_species:

		# Update the selection dict
		selection_dict['IN'][config_data.returnColumnPath('Species')] = include_species

	# Check if species exclude is defined
	if exclude_species:

		# Update the selection dict
		selection_dict['NOT IN'][config_data.returnColumnPath('Species')] = exclude_species

	# Check if genus include is defined
	if include_genus:

		# Update the selection dict
		selection_dict['LIKE'][config_data.returnColumnPath('Species')] = include_genus

	# Check if genus exclude is defined
	if exclude_genus:

		# Update the selection dict
		selection_dict['NOT LIKE'][config_data.returnColumnPath('Species')] = exclude_genus

	# Check if genus include is defined
	if include_nests:

		# Update the selection dict
		selection_dict['IN'][config_data.returnColumnPath('From Nest?')] = ['Yes']

	# Check if genus exclude is defined
	if exclude_nests:

		# Update the selection dict
		selection_dict['NOT IN'][config_data.returnColumnPath('From Nest?')] = ['Yes']

	# Update log
	logging.info('Successfully assigned selection statement dict(s)')

	return selection_dict

'''


import os
import sys
import csv
import logging

from kocher_tools.common import readCommonFile
from kocher_tools.database import insertValues, updateValues, retrieveValues, confirmValue

def convertPlateWell (cursor, table, key, plate, well):

	# Get the key data from the Storage table
	key_data = retrieveValues(cursor, [table], {'IN':{'Plate':[plate], 'Well':[well]}}, [key])

	# Check if more than one ID was found
	if len(key_data) > 1: 
		raise Exception('Unable to assign %s' % key)

	# Check if no IDs were found
	if not key_data:
		raise Exception('Unable to find %s-%s' % (plate, well))

	# Update log
	logging.info('Plate/Well conversion successful')

	return key_data[0][0]

def addStorageFileToDatabase (cursor, table, storage_file):

	# Update log
	logging.info('Uploading storage file (%s) to database' % storage_file)

	# Loop the storage file by line
	for header, storage_data in readCommonFile(storage_file):

		# Insert the loc into the database
		insertValues(cursor, table, header, storage_data)

	# Update log
	logging.info('Upload successful')

def updateStorageFileToDatabase (cursor, table, select_key, storage_file):

	# Update log
	logging.info('Uploading storage file (%s) to database' % storage_file)

	# Loop the storage file by line
	for header, storage_data in readCommonFile(storage_file):

		# Check if the selection key isn't among the headers
		if select_key not in header:
			raise Exception('Selection key (%s) not found. Please check the input file' % select_key)

		# Create an empty string to store the select_value
		select_value = ''

		# Create an empty set dict
		storage_set_dict = {}

		# Create an empty selection dict
		storage_select_dict = {}
		storage_select_dict['IN'] = {}

		# Loop the header ans sample data
		for header_column, storage_value in zip(header, storage_data):

			# Check if the current column is the selection key
			if header_column == select_key:

				# Update the select value
				select_value = storage_value

				# Populate the selection dict
				storage_select_dict['IN'][header_column] = [storage_value]

			else:

				# Populate the selection dict
				storage_set_dict[header_column] = storage_value

		# Check that selected value is present
		if not confirmValue(cursor, table, select_key, select_value):

			# If not, log warning
			logging.warning('Entry (%s) not found, unable to update. Please check input' % select_value)

		# Update the values for the selected value
		updateValues(cursor, table, storage_select_dict, storage_set_dict)

	# Update log
	logging.info('Upload successful')



import io
import kocher_tools
from setuptools import setup

with io.open("README.rst", "rt", encoding="utf8") as f:
    readme = f.read()

# List of required non-standard python libraries
requirements = ['pyyaml',
                'pandas',
                'Biopython',
                'sqlalchemy',
                'gffutils',
                'openpyxl',
                'tox']

# Executable scripts in the package
tool_scripts = ['kocher_tools/barcode_pipeline.py',
                'kocher_tools/barcode_filter.py',
                'kocher_tools/create_database.py',
                'kocher_tools/insert_file.py',
                'kocher_tools/gff_position_stats.py',
                'kocher_tools/gff_chrom_stats.py',
                'kocher_tools/gff_add_features.py']

setup(name=kocher_tools.__title__,
      version=kocher_tools.__version__,
      project_urls={"Documentation": "https://kocher-guides.readthedocs.io/",
                    "Code": "https://github.com/kocherlab/kocher_tools",
                    "Issue tracker": "https://github.com/kocherlab/kocher_tools/issues"},
      license=kocher_tools.__license__,
      url=kocher_tools.__url__,
      author=kocher_tools.__author__,
      author_email=kocher_tools.__email__,
      maintainer="Andrew Webb",
      maintainer_email="19213578+aewebb80@users.noreply.github.com",
      description=kocher_tools.__summary__,
      long_description=readme,
      include_package_data=True,
      packages=['kocher_tools'],
      package_data={'kocher_tools': ['data/*.txt']},
      install_requires=requirements,
      scripts=tool_scripts,
      python_requires=">=3.6")



#!/usr/bin/env python
import os
import sys
import csv
import argparse
import logging
import json
import copy

import pandas as pd
from collections import defaultdict

from kocher_tools.logger import startLogger

def barcodeFilterParser ():
	'''
	Barcode Filter Parser

	Assign the parameters for the barcode filter.

	Parameters
	----------
	sys.argv : list
		Parameters from command lind

	Raises
	------
	IOError
		If the specified files do not exist
	'''

	def confirmFile ():
		'''Custom action to confirm file exists'''
		class customAction(argparse.Action):
			def __call__(self, parser, args, value, option_string=None):
				if not os.path.isfile(value):
					raise IOError('%s not found' % value)
				setattr(args, self.dest, value)
		return customAction

	def metavarList (var_list):
		'''Create a formmated metavar list for the help output'''
		return '{' + ', '.join(var_list) + '}'

	filter_parser = argparse.ArgumentParser(formatter_class = argparse.ArgumentDefaultsHelpFormatter)

	# Input file args
	filter_parser.add_argument('--blast-file', help = 'Filename of the BLAST output', type = str, action = confirmFile(), required = True)

	# Control filters
	filter_parser.add_argument('--negative-control', help = 'Negative control ID', type = str, nargs = '+')
	filter_parser.add_argument('--negative-control-file', help = 'File of negative control IDs', type = str, action = confirmFile())

	# Basic filter cutoff args\
	filter_parser.add_argument('--abundance-cutoff', help = 'Read abundance Cutoff', type = int, default = 25)
	filter_parser.add_argument('--evalue-cutoff', help = 'E-Value Cutoff', type = float, default = 0.00001)
	filter_parser.add_argument('--coverage-cutoff', help = 'Coverage cutoff - i.e. percentage of sequence aligned', type = float, default = 0.90)
	filter_parser.add_argument('--identity-cutoff', help = 'Identity cutoff - i.e. identity percentage between alignment sequences', type = float, default = 0.95)

	# Best hit args
	sort_by = ('percent_ident', 'align_len')
	filter_parser.add_argument('--sort-best-hits-by', metavar = metavarList(sort_by), help = 'Sort method of best hits', choices = sort_by, default = ['percent_ident'], nargs = '+', type = str)
	filter_parser.add_argument('--dont-merge-species', help = 'Do not merge best hits within a single species', action = 'store_true')

	# Output file args
	filter_parser.add_argument('--out-prefix', help = 'Output prefix for the BLAST file', type = str, default = 'Filtered_')
	filter_parser.add_argument('--out-blast', help = 'Output filename for the BLAST file. Overrides --out-prefix', type = str)
	filter_parser.add_argument('--out-failed', help = 'Output filename for BLAST failures json file', type = str, default = 'barcode_failures.json')
	filter_parser.add_argument('--out-log', help = 'Filename of the log file', type = str, default = 'barcode_filter.log')
	filter_parser.add_argument('--overwrite', help = 'Overwrite previous output', action = 'store_true')

	# Return the arguments
	return filter_parser.parse_args()

def yieldBestHits (blast_reader):

	# Create a string to store the query being filtered
	filter_query = None

	# Create a float to store the lowest e-value
	filter_lowest_evalue = None

	# Create list to store the best hits of the current query
	filter_best_hits = []

	# Read the file, line by line
	for blast_dict in blast_reader:

		# Save the current query ID
		current_query = blast_dict['Query ID']

		# Save the current evalue
		current_evalue = float(blast_dict['E-Value'])

		# Check if the current query doesn not match the filter query
		if filter_query != current_query:

			# Check if any best hits have been assigned
			if filter_best_hits:

				yield filter_query, filter_best_hits

			# Assign the new filter query
			filter_query = current_query

			# Assign the new lowest e-value
			filter_lowest_evalue = current_evalue

			# Assign the new best hit
			filter_best_hits = [blast_dict]

		else:

			# Check if the current e-value is lower than the filtered e-value
			if filter_lowest_evalue > current_evalue:

				# Assign the new lowest e-value
				filter_lowest_evalue = current_evalue

				# Assign the new best hit
				filter_best_hits = [blast_dict]

			# Check if the current e-value is equal to the filtered e-value
			elif filter_lowest_evalue == current_evalue:

				# Append the best blast hit
				filter_best_hits.append(blast_dict)

	# Check if there is data after the loop is finished
	if filter_query and filter_best_hits:

		# Yield the data
		yield filter_query, filter_best_hits

def sortBestHits (best_hits, method):

	# Create list to store the sorted best hits
	sorted_best_hits = []

	# Check if the method is sort by percent identity
	if method == 'percent_ident':

		# Create a float to store the highest percent identity
		best_percent_identity = 0.0

		# Loop the filtered hits
		for blast_hit in best_hits:

			# Save the current identity
			current_identity = float(blast_hit['Percent Identity'])

			# Check if this hit has a higher percent identity
			if best_percent_identity < current_identity:

				# Update the best hit
				sorted_best_hits = [blast_hit]

				# Update the highest percent identity
				best_percent_identity = current_identity

			# Check if this hit has a higher percent identity
			elif best_percent_identity == current_identity:

				# Update the best hit
				sorted_best_hits.append(blast_hit)

	# Check if the method is sort by percent identity
	elif method == 'align_len':

		# Create a float to store the highest alignment length
		best_align_len = 0.0

		# Loop the filtered hits
		for blast_hit in best_hits:

			# Save the current alignment length
			current_align_len = float(blast_hit['Alignment Length'])

			# Check if this hit has a higher alignment length
			if best_align_len < current_align_len:

				# Update the best hit
				sorted_best_hits = [blast_hit]

				# Update the highest alignment length
				best_align_len = current_align_len

			# Check if this hit has a higher alignment length
			elif best_align_len == current_align_len:

				# Update the best hit
				sorted_best_hits.append(blast_hit)

	# Return the sorted best hits
	return sorted_best_hits

def main():

	# Assign the barcode args
	barcode_args = barcodeFilterParser()

	# Create the log file
	startLogger(barcode_args.out_log)

	# Check if a BLAST output filename was defined
	if barcode_args.out_blast:
		
		# Define the BLAST output filename
		blast_out_filename = barcode_args.out_blast

		# Get the output dirname
		out_dirname = os.path.dirname(blast_out_filename)

		# Define the JSON output filename
		failed_out_filename = os.path.join(out_dirname, barcode_args.out_failed)

	# If not, define using a prefix
	else:

		# Get the BLAST basename
		blast_basename = os.path.basename(barcode_args.blast_file)

		# Get the BLAST dirname
		blast_dirname = os.path.dirname(barcode_args.blast_file)

		# Define the BLAST output filename
		blast_out_filename = os.path.join(blast_dirname, barcode_args.out_prefix + blast_basename)

		# Define the JSON output filename
		failed_out_filename = os.path.join(blast_dirname, barcode_args.out_failed)

	# Check if previous output should be overwritten
	if barcode_args.overwrite:

		# Remove the previous output, if it exists
		if os.path.exists(blast_out_filename):
			os.remove(blast_out_filename)
		if os.path.exists(failed_out_filename):
			os.remove(failed_out_filename)

	# Check if previous output shouldn't be overwritten
	else:

		# Check if previous output exists
		if os.path.exists(blast_out_filename) or os.path.exists(failed_out_filename):

			# Raise an exception
			raise Exception('Output already exists. Please alter the output arguments or use --overwrite')

	# Create a list of negative control IDs
	negative_control_list = []

	# Check if negative control IDs were given on the command line
	if barcode_args.negative_control:

		# Assign the negative control IDs
		negative_control_list.extend(barcode_args.negative_control)

	# Check if a file of negative control IDs was given
	if barcode_args.negative_control_file:

		# Open the file
		with open(barcode_args.negative_control_file, 'r') as negative_control_file:

			# Read the file, line by line
			for negative_control_line in negative_control_file:

				# Assign the negative control ID
				negative_control_list.append(negative_control_line.strip())

	# Create a list to store failed sample dictonaries
	failed_samples_list = []

	# Open the BLAST input file
	with open(barcode_args.blast_file) as blast_file:

		# Read the blast using DictReader
		blast_reader = csv.DictReader(blast_file, delimiter = '\t')

		# Create the blast output file
		blast_out_file = open(blast_out_filename, 'w')

		# Create the blast writer using DictReader
		blast_writer = csv.DictWriter(blast_out_file, fieldnames = blast_reader.fieldnames, delimiter = '\t')

		# Write the header for the output file
		blast_writer.writeheader()

		# Loop each query with its best hits
		for current_query, best_blast_hits in yieldBestHits(blast_reader):

			# Create a list to store the filtered blast hits
			filtered_blast_hits = []

			# Save the current ID
			current_ID = current_query.rsplit('_', 1)[0]

			# Save the current abundance
			current_abundance = int(current_query.split('=')[1])

			# Check if an abundance cutoff was specified
			if barcode_args.abundance_cutoff:

				# Check if the current abundance is below the cutoff
				if current_abundance < barcode_args.abundance_cutoff:

					# Save warning message
					warning_message = 'Insufficent read abundance'
					
					# Create dict to hold all relevant information for the failed sample
					failed_sample_dict = {'Query ID':current_query, 'Status':'Insufficent reads'}

					# Add the dict to the list
					failed_samples_list.append(failed_sample_dict)

					# Log the failure
					logging.warning('%s: %s' % (current_query.split(';')[0], warning_message))

					continue

			# Loop each best hit
			for blast_hit in best_blast_hits:

				# Save the current evalue
				current_evalue = float(blast_hit['E-Value'])

				# Save the current query length
				current_query_length = float(blast_hit['Query Length'])

				# Save the current alignment length
				current_alignment_length = float(blast_hit['Alignment Length'])

				# Save the current identity
				current_identity = float(blast_hit['Percent Identity'])

				# Check if an E-Value cutoff was specified
				if barcode_args.evalue_cutoff:

					# Check if the current evalue is larger than the cutoff
					if current_evalue > barcode_args.evalue_cutoff:
						continue
						
				# Check if a coverage cutoff was specified
				if barcode_args.coverage_cutoff:

					# Save the percent coverage
					current_coverage = current_alignment_length / current_query_length

					# Check if the current coverage is samller than the cutoff
					if current_coverage < barcode_args.coverage_cutoff:
						continue

				# Check if an identity cutoff was specified
				if barcode_args.identity_cutoff:

					# Check if the current identity is smaller than the cutoff
					if current_identity < barcode_args.identity_cutoff:
						continue

				# Add the blast hit to the filtered list, if passed
				filtered_blast_hits.append(blast_hit)

			# Check if no blast hits passed the filter
			if len(filtered_blast_hits) == 0:

				# Save warning message
				warning_message = 'Filtered resulted in the removal of all data'
				
				# Create dict to hold all relevant information for the failed sample
				failed_sample_dict = {'Query ID':current_query, 'Status':'No Hits'}

				# Add the dict to the list
				failed_samples_list.append(failed_sample_dict)

				# Log the failure
				logging.warning('%s: %s' % (current_query.split(';')[0], warning_message))

			# Copy the filtered list before sorting
			sorted_best_hits = copy.deepcopy(filtered_blast_hits)

			# Loop the sort methods
			for sort_method in barcode_args.sort_best_hits_by:

				# Sort the best hits
				sorted_best_hits = sortBestHits(sorted_best_hits, sort_method)

			# Check if there is more than one sorted best hit
			if len(sorted_best_hits) > 1:

				# Create a list for checking the number of species
				sorted_species_list = []

				# Create a list for checking the number of bins
				sorted_bin_list = []

				# Loop the sorted best hits
				for sorted_best_hit in sorted_best_hits:

					# Assign the species of the hit
					sorted_species = sorted_best_hit['Subject ID'].split('|')[1].replace('_', ' ')

					# Append the species to the list
					sorted_species_list.append(sorted_species)

					# Assign the bin of the hit
					sorted_bin = sorted_best_hit['Subject ID'].split('|')[2]

					# Append the bin to the list
					sorted_bin_list.append(sorted_bin)

				# Remove duplicate species
				sorted_species_list = list(set(sorted_species_list))

				# Remove duplicate bins
				sorted_bin_list = list(set(sorted_bin_list))

				# Check if a BOLD:N/A was found
				if 'BOLD:N/A' in sorted_bin_list and len(sorted_bin_list) > 1:

					# Move the BOLD:N/A to the end of the list
					sorted_bin_list.append(sorted_bin_list.pop(sorted_bin_list.index('BOLD:N/A'))) 

				# Check if more than a single species among the best hits
				if len(sorted_species_list) > 1:

					# Save warning message
					warning_message = 'Multiple species identified'
					
					# Create dict to hold all relevant information for the failed sample
					failed_sample_dict = {'Query ID':current_query, 'Status':'Ambiguous Hits', 
					                      'Species':sorted_species_list , 'Bins':sorted_bin_list}

					# Add the dict to the list
					failed_samples_list.append(failed_sample_dict)

					# Log the failure
					logging.warning('%s: %s' % (current_query.split(';')[0], warning_message))

					continue

				# Check if merging should occur, and if there is only a single species among the best hits
				elif not barcode_args.dont_merge_species and len(sorted_species_list) == 1:

					# Take the first sorted hit, update this later with preferences
					sorted_best_hits = [sorted_best_hits[0]]

			# Check if any negative control IDs are assigned
			if sorted_best_hits and negative_control_list and current_ID in negative_control_list:

				logging.warning('Negative control (%s) passed filters. Please consider stricter cutoffs' % current_ID)

				continue

			# Loop the sorted best hits
			for sorted_best_hit in sorted_best_hits:

				# Write the passed blast entry to the output
				blast_writer.writerow(sorted_best_hit)

		# Close the output file
		blast_out_file.close()

	# Check if any samples failed to find best hits
	if failed_samples_list:

		logging.warning('Samples (%s) failed filters. Samples may be found in: %s' % (len(failed_samples_list), failed_out_filename))

	# Open the json failure file
	with open(failed_out_filename, 'w') as json_failed_file:

		# Dump the failed sample list to the JSON file
		json.dump(failed_samples_list, json_failed_file, indent = 4)

if __name__== "__main__":
	main()